import functools
import json


class RouteRegistry:
    """
    A client-side imitation of a Web Framework view path registry.

    This class serves to create backwards compatibility and fallback capable code
    between established REST API calls and client-side data access objects (DAOs).
    """

    routes = {}

    def register(self, method, route):
        """
        Register a class method to a specific route.

        >>> @registry.register('GET', '{var}/path')
        >>> def my_class_method(self, var):
        >>>     print(var)

        :param method: The Http method (GET, POST, PUT, PATCH or DELETE).
        :param route: The unique route to associate with the class method. Supports multiple dynamic path variables.
        :returns: Calls the respective method and returns its result.
        """

        def decorator(func):

            @functools.wraps(func)
            def wrapped(instance, path, *args, **kwargs):
                variables = []
                route_parts = route.split('/')
                path_parts = path.split('/')

                # Grab the variables from the requested path
                for route_part, path_part in zip(route_parts, path_parts):
                    if route_part.startswith('{') and route_part.endswith('}'):
                        variables.append(path_part)

                kwargs = self._load_json(kwargs)

                return func(instance, *variables, *args, **kwargs)

            # Register the route to the registry
            if method not in self.routes: self.routes[method] = {}
            self.routes[method][route] = wrapped

            return wrapped

        return decorator

    def call_method(self, instance, http_method: str, path: str, *args, **kwargs):
        """
        Runs the wrapped function that corresponds to the given path.

        :param instance: Reference to the class instance
        :param http_method: The Http method (GET, POST, PUT, PATCH or DELETE)
        :param path: The requested path
        :param args: Any additional arguments to be passed to the function
        :param kwargs: Any additional keyword arguments to be passed to the function
        :returns: The wrapped function if path matches.
        :raises ValueError: If no match was found for given path.
        """
        if http_method not in self.routes:
            raise ValueError(f"No matching route found for ({http_method}, {path}).")

        path_parts = list(filter(None, path.split('/')))

        func = self.match_specific_route(self.routes[http_method], path_parts) \
               or self.match_any_route(self.routes[http_method], path_parts)

        if func: return func(instance, path, *args, **kwargs)

        raise ValueError(f"No matching route found for ({http_method}, {path}).")

    @staticmethod
    def _load_json(kwargs: dict):
        """
        Convert JSON compatible strings back to the original object.

        :param kwargs: Dictionary with all keyword arguments.
        :return: Dictionary with the correct objects loaded.
        """
        for key, item in kwargs.items():
            if isinstance(item, str):
                try:
                    kwargs[key] = json.loads(item)
                except ValueError:
                    pass
        return kwargs

    @staticmethod
    def match_specific_route(routes: dict, path_parts: list):
        """
        Match with specific route (i.e., {var}/path).

        :param routes: Dictionary with all the routes of a given Http method
        :param path_parts: The list of path parts to compare with the registry path parts
        :return: The wrapped function if paths match, otherwise None
        """
        for reg_path, func in routes.items():
            reg_path_parts = reg_path.split('/')
            if len(reg_path_parts) == len(path_parts):
                match = []
                for part, user_part in zip(reg_path_parts, path_parts):
                    if not part.startswith('{') and not part.endswith('}'):
                        match.append(True if part == user_part else False)
                if match and all(match):
                    return func
        return None

    @staticmethod
    def match_any_route(routes: dict, path_parts: list):
        """
        Get the first route that matches (i.e., {var1}/{var2}).

        :param routes: Dictionary with all the routes of a given Http method
        :param path_parts: The list of path parts to compare with the registry path parts
        :return: The wrapped function if paths match, otherwise None
        """
        for reg_path, func in routes.items():
            reg_path_parts = reg_path.split('/')
            if len(reg_path_parts) == len(path_parts):
                matches = True
                for part, user_part in zip(reg_path_parts, path_parts):
                    if not part.startswith('{') and not part.endswith('}') and part != user_part:
                        matches = False
                        break
                if matches:
                    return func
        return None



from collections.abc import Iterable

__all__ = ['ToriiService']

import sys
import json as jsonlib

from torii.services.service import Service
from torii.data import ToriiObject, BusinessObject
from torii.data import Struct
from torii.exception import ToriiException


class ToriiService(Service):

    def __init__(self, torii, name, base_path=None, object_class=ToriiObject):
        """
        Create a Gateway service serving ToriiObjects
        """
        Service.__init__(self, torii, name, base_path)

        self._object_class = object_class

        self._torii = torii

    def action(self, action=None, obj=None, params=None,
               update=True,
               send_payload=True,
               raw_result=False,
               cast=None,
               service=None):
        """
        Simplified posting action
        """
        if not action or not obj:
            raise ToriiException('Missing parameters')

        # apply the action on all the elements of a sequence independently
        if not isinstance(obj, self._object_class) and isinstance(obj, Iterable):
            return [self.action(action, obji, params) for obji in obj]

        if not isinstance(obj, self._object_class):
            raise ToriiException('Wrong parameter obj, obj parameter type {0} expected'.format(self._object_class))

        if obj is not None:
            path = '{0}/{1}'.format(obj.id, action)
            # prepare the payload if the action requires the object content
            payload = obj.to_json() if send_payload else None
        else:
            path = action
            payload = None

        result = self.request('POST',
                              path=path,
                              json=payload,
                              params=params,
                              raw_result=raw_result)
        if raw_result:
            return result
        elif update:
            return obj.from_json(result)
        elif cast and issubclass(cast, ToriiObject):
            return cast(json=result, service=service if service else None)
        else:
            return Struct(json=result)

    def index(self, fields=None, default=False, unique=False):
        params = {
            'default': default,
            'unique': unique
        }

        if fields is None:
            fields = []

        return self.request_put(path='index', json=fields, params=params)

    def new(self, create=True, **kwargs):
        """
        Initialize a new shell object
        Be careful this object is not created in Gateway
        :return: the new object
        """
        try:
            params = {'create': create}
            if kwargs:
                for key, value in kwargs.items():
                    if isinstance(value, ToriiObject):
                        params[key] = value.id
                    else:
                        params[key] = value

            json = self.request_get(path='new', params=params)
            obj = self._object_class(json, self)

            return obj

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to instanciate a shell for {0}\n{1}'
                                 .format(self._object_class, e)).with_traceback(sys.exc_info()[2])

    def create(self, obj):
        """
        Register an object in torii
        :param obj: the object to register
        :return: the modified object
        """
        try:
            if isinstance(obj, self._object_class):
                json = self.request_post(json=obj.to_json())
                # update current object
                obj.from_json(json)
                obj._service = self
                return obj
            elif isinstance(obj, list):
                return self.request_post(path='bulk', json=obj)
            else:
                json = self.request_post(json=obj)
                # create a new ToriiObject
                return self._object_class(json, self)

        except ToriiException as e:
            raise e
        except Exception as e:
            raise ToriiException('Failed to create a {0} in the system\n{1}'
                                 .format(self._object_class, e)).with_traceback(sys.exc_info()[2])

    def list(self, dir='ASC', sort=None, limit=100, start=0, timeout=60, filters=None, strict_filter=False,
             include=None, no_paging=False, count=False, raw=False, cast=True, **kwargs):
        """
        List objects from the service
        :param dir: the sorting direction
        :param sort: the sorting property
        :param timeout: the request timeout in seconds
        :param limit: the maximum number of objects to be listed
        :param start: the index of the first object to be listed (regarding the sorting and filters)
        :param filters: the filters to apply
        :param strict_filter: filter the strings with 'eq' instead of 'like'
        :param include: the fields to include when getting the objects
        :param no_paging: use raw listing (no paging, no filtering)
        :param count: return the total count of elements along with the page
        :param raw: don't cast the objects in Java
        :param kwargs: quick filters

        :return: a list of objects
        """
        try:
            # mandatory if you don't want to pollute the default value and so the following calls
            # sometimes I hate python's guts !!!
            if filters is None:
                filters = []

            # add filters corresponding to the free arguments
            for key, value in kwargs.items():
                if key == 'id':
                    filters.append({'property': '_id', 'value': value, 'filterType': 'string', 'operator': 'eq'})
                elif type(value) in (float, int):
                    filters.append({'property': key, 'value': value, 'filterType': 'number', 'operator': 'eq'})
                elif isinstance(value, ToriiObject):
                    filters.append({'property': key, 'value': value.id, 'filterType': 'string', 'operator': 'eq'})
                elif strict_filter:
                    filters.append({'property': key, 'value': value, 'filterType': 'string', 'operator': 'eq'})
                else:
                    filters.append({'property': key, 'value': value, 'filterType': 'string', 'operator': 'like'})

            # request the corresponding page
            if not no_paging:
                params = {
                    'dir': dir,
                    'timeout': timeout,
                    'sort': sort,
                    'limit': limit,
                    'start': start,
                    'filter': jsonlib.dumps(filters),
                    'include': include,
                    'count': count,
                    'raw': raw
                }

                if sort:
                    params['sort'] = sort
                    params['limit'] = limit

                page = self.request_get(path='page', params=params)

                # convert the json as an object list
                if cast:
                    page_content = [self._object_class(json, self) for json in page['result']]
                else:
                    page_content = page['result']

                if count:
                    return page_content, page['size']
                else:
                    return page_content

            else:
                list = self.request_get()
                return [self._object_class(json, self) for json in list]

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to list {0}s from Gateway\n{1}'
                                 .format(self._object_class, e)).with_traceback(sys.exc_info()[2])

    def count(self, filters=None, strict_filter=False, **kwargs):
        """
        List objects from the service
        :param filters: the filters to apply
        :param strict_filter: filter the strings with 'eq' instead of 'like'
        :param kwargs: quick filters
        """

        # mandatory if you don't want to pollute the default value and so the following calls
        # sometimes I hate python's guts !!!
        if filters is None:
            filters = []

        # add filters corresponding to the free arguments
        for key, value in kwargs.items():
            if key == 'id':
                filters.append({'property': '_id', 'value': value, 'filterType': 'string', 'operator': 'eq'})
            elif type(value) in (float, int):
                filters.append({'property': key, 'value': value, 'filterType': 'number', 'operator': 'eq'})
            elif isinstance(value, ToriiObject):
                filters.append({'property': key, 'value': value.id, 'filterType': 'string', 'operator': 'eq'})
            elif strict_filter:
                filters.append({'property': key, 'value': value, 'filterType': 'string', 'operator': 'eq'})
            else:
                filters.append({'property': key, 'value': value, 'filterType': 'string', 'operator': 'like'})

        return self.request_get(path='count', params={'filter': jsonlib.dumps(filters)})

    def get(self, id=None, strict_filter=True, take_first=False, include=None, cast=True, **kwargs):
        """
        Get a single object
        :param id: the id of the object to get
        :param strict_filter: filter the strings with 'eq' instead of 'like'
        :param take_first: if several objects match the filters, take the first one
        :param kwargs: quick filters
        
        :return:
        """

        if id:
            params = {
                'include': include,
            }
            # get by id
            json = self.request_get(path=id, params=params)
            if json is not None:
                if cast:
                    return self._object_class(json, service=self)
                else:
                    return json
            else:
                return None

        else:
            # get by filter
            list = self.list(limit=10, strict_filter=strict_filter, include=include, cast=cast, **kwargs)

            if len(list) == 0:
                return None
            elif len(list) == 1 or take_first:
                return list[0]
            else:
                raise ToriiException('Several {0} match the filters (>= {1})'.format(self._object_class, len(list)))

    def refresh(self, obj):
        """
        Get the latest information from the server
        :param obj: the object to refresh
        :return:
        """
        if not isinstance(obj, self._object_class):
            raise ToriiException('Bad parameter, obj must be a {0}'.format(self._object_class.__name__))

        try:
            json = self.request_get(path=obj.id)
            obj.from_json(json)

            return obj
        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to refresh {0} in the system\n{1}'
                                 .format(self._object_class, e)).with_traceback(sys.exc_info()[2])

    def update(self, obj, create=False, **kwargs):
        """
        Update an object
        :param obj:
        :return:
        """

        if create:
            try:
                if self.get(obj.id) is None:
                    return self.create(obj, **kwargs)
            except:
                return self.create(obj, **kwargs)

        try:
            if isinstance(obj, self._object_class):
                if obj.has_changed():
                    # update the remote version
                    json = self.request_post(path=obj.id, json=obj.to_json(), params=kwargs)
                else:
                    # update local version
                    json = self.request_get(path=obj.id)
            elif isinstance(obj, dict) and '_id' in obj:
                # update the remote version
                json = self.request_post(path=obj['_id'], json=obj, params=kwargs)
            elif isinstance(obj, list):
                obj_json = []
                for o in obj:
                    if isinstance(o, self._object_class):
                        obj_json.append(o.to_json())
                    elif isinstance(o, dict):
                        obj_json.append(o)
                    else:
                        raise ToriiException("unexpected case in list update: {}".format(type(o)))
                json = self.request_put(path='bulk', json=obj_json, params=kwargs)
                if json:
                    obj = []
                    for j in json:
                        obj.append(BusinessObject(json=j))
                else:
                    obj = json

                return obj
            else:
                raise ToriiException('unexpected case: please think about it before removing tis exception :-)')
                json = self.request_post(path=obj['_id'], json=obj, params=kwargs)

            if isinstance(obj, self._object_class):
                obj.from_json(json)
            else:
                obj = json

            return obj

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to update {0}\n{1}'.format(obj, e)).with_traceback(sys.exc_info()[2])

    def patch(self, id, operations):
        """
        Patch the document with given id by applying the given operations
        :param id: the id of the document
        :param operations: expected format {"op": "replace", "key": "mykey", "value": "myvalue"}
        :return:
        """
        return self.request_patch(path=id, json=operations)

    def patches(self, identified_operations=None, cast=True, **kwargs):
        """
        Patch the documents identified by applying their associated operations
        :param identified_operations: expected format {"myid": {"op": "replace", "key": "mykey", "value": "myvalue"}, }
        :param cast: cast to object or return raw json
        :return:
        """
        if identified_operations is None:
            raise ToriiException("no operation provided for patch")

        json = self.request_patch(path='', json=identified_operations, params=kwargs)
        if json:
            if cast:
                obj = [self._object_class(j, self) for j in json]
            else:
                obj = json
        else:
            obj = json
        return obj

    def delete(self, obj, force=False):
        """
        Delete an object
        :param obj: the object to refresh
        :param force: force the delete
        :return:
        """

        try:
            if isinstance(obj, self._object_class):
                self.request_delete(path=obj.id, params={"force": force})

                return obj
            elif isinstance(obj, list):
                params = {
                    'filter': jsonlib.dumps(obj),
                    'force': force
                }

                self.request_delete(path='', params=params)
            else:
                raise ToriiException('Bad parameter, obj must be a {0}'.format(self._object_class.__name__))

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to delete {0} from the system\n{1}'
                                 .format(self._object_class, e)).with_traceback(sys.exc_info()[2])

    def distinct(self, property, limit=None, filters=None, stats=None):
        """"""
        try:
            params = {}
            if limit is not None:
                params['limit'] = limit
            if filters is not None:
                params['filter'] = jsonlib.dumps(filters)
            if stats is not None:
                params['stats'] = jsonlib.dumps(stats)

            values = self.request_get(path='distinct/' + property, params=params)

            return Struct(values)
        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to get distinct values for property {0} on {1}\n{2}'
                                 .format(property, self._object_class, e)).with_traceback(sys.exc_info()[2])




from torii.data.struct import Struct

class Record(Struct):
    """
    Wrapper for the task records
    """

    def __init__(self, json={}, service=None):
        Struct.__init__(self, json)

        self._service = service



import sys

if sys.version_info >= (3, 0):
    class ToriiException(Exception):
        pass
else:
    class ToriiException(Exception):

        def with_traceback(self, x):
            return (self, None, x)



import time
import webbrowser

from torii.exception import ToriiException
from torii.services.torii_service import ToriiService
from torii.data import Task, Picom


class TaskService(ToriiService):

    def __init__(self, torii):
        """
        Create the task service
        """
        ToriiService.__init__(self, torii=torii, name='tasks', object_class=Task)

    def refresh(self, task):
        task = super(TaskService, self).refresh(task)

        return task

    def new(self, create=True, **kwargs):
        # by default, create the task in the current container
        if 'parent' not in kwargs and self._torii.current_container:
            kwargs['parent'] = self._torii.current_container.id

        return ToriiService.new(self, create=create, **kwargs)


    def create(self, task=None, template=None):

        if task is None :
            if template is not None:
                if not isinstance(template, Task):
                    # try to find the template by name
                    template = self.get(name=template, status='TEMPLATE')

                task = Task(template)
                task.status = 'UNKNOWN'
            else:
                raise ToriiException('Cannot create task without template')

        # use the default container if necessary/possible
        if (not hasattr(task, 'parent') or not task.parent) and self._torii.current_container:
            task.parent = self._torii.current_container.id

        return ToriiService.create(self, task)


    def validate(self, task):
        """
        Validate a task

        :param task: the task to be validated
        :return:
        """
        data = {
            'task': task.to_json(),
            'save': True
        }

        if task.has_changed():
            data['previousTask'] = task.orig

        json = self.request('POST', path='{0}/validate'.format(task.id), json=data)
        task.from_json(json)
        return task


    def submit(self, task):
        """
        Submit a task

        :param task: the task to submit
        :return:
        """
        if task.has_changed():
            self.update(task)

        return self.action('submit', task)

    def lowla(self, task, **kwargs):
        """
        Submit a Low Latency task on a running LowLa Engine

        :param task: the task to submit
        :return:
        """

        return self.action('lowla', task, send_payload=True, update=False, cast=Task, params=kwargs)


    def clone(self, task, parent=None):
        """
        Clone a task

        :param task: the task to clone
        :return:
        """

        if task.has_changed():
            self.update(task)

        # by default, create the task in the current container
        if parent :
            params = {'parentId': parent.id}
        else:
            params = {'parentId': self._torii.current_container.id}

        clone = self.action(action='clone', obj=task,
                            send_payload=False, update=False,
                            params=params,
                            cast=Task, service=self)

        # refresh the task (its cloned counter have changed)
        task.refresh()

        return clone

    def picomize(self, task):
        """
        Picomize a hierarchical task

        :param task: the task to picomize
        :return:
        """
        if task.has_changed():
            self.update(task)

        picom = self.action(action='picomize', obj=task,
                            send_payload=False, update=False,
                            cast=Picom, service=self)

        return picom

    def command(self, task, commandId, commandOption=None):
        """
        Throw a task command

        :param task: the task
        :param commandId: the command type
        :param commandOption: the options of the command if any
        :return:
        """

        if task.has_changed():
            self.update(task)

        params = {'commandId': commandId, 'commandOption': commandOption}
        return self.action('taskCommand', task, params=params, send_payload=False, update=False)


    def job_command(self, task, jobIdx, commandId, commandOption=None):
        """
        Throw a job command

        :param task: the task
        :param jobIdx: the job index
        :param commandId: the command type
        :param commandOption: the options of the command if any
        :return:
        """

        if task.has_changed():
            self.update(task)

        params = {'jobId': task.jobs[jobIdx]._id, 'commandId': commandId, 'commandOption': commandOption}
        return self.action('jobCommand', task, params=params, send_payload=False, update=False)


    def job_command(self, task, jobIdx, commandId, commandOption=None):
        """
        Throw a job command

        :param task: the task
        :param jobIdx: the job index
        :param commandId: the command type
        :param commandOption: the options of the command if any
        :return:
        """

        if task.has_changed():
            self.update(task)

        params = {'jobId': task.jobs[jobIdx]._id, 'commandId': commandId, 'commandOption': commandOption}
        return self.action('jobCommand', task, params=params, send_payload=False, update=False)

    def kill(self, task):
        return self.command(task, 'kill')

    def pause(self, task):
        return self.command(task, 'hold')

    def resume(self, task):
        return self.command(task, 'release')


    def get_all_status(self, task=None, parent=None, root=None, status=None):
        if isinstance(task, str):
            tasks = [self.get(id)]
        elif isinstance(task, Task):
            task.refresh()
            tasks = [task]
        elif isinstance(parent, (Task, str)):
            tasks = self.list(parent=parent)
        elif isinstance(root, (Task, str)):
            tasks = self.list(root = root)
        elif self._torii.current_container:
            tasks = self.list(parent=self._torii.current_container)
        else:
            raise ToriiException('Bad parameter, check_active requires task, parent or root being Task or str')

        return [task.status for task in tasks]


    def check_status(self, task=None, parent=None, root=None, status=None):
        """
        Check if a task or the children or the descendants are active

        by default, if no argument is provided the service will check the children of the current container

        :param task: the task to check (id or Task)
        :param parent: the parent of the tasks to check (id or Task)
        :param root: the root of the tasks to check (id or Task)
        :param status: the status against which the task(s) is checked
        :return: True or False
        """

        stati = self.get_all_status(self, task, parent, root, status)
        return all([Task.is_status(s, status) for s in stati])


    def check_active(self, task=None, parent=None, root=None, status=None):
        """
        Check if a task or a group of tasks are active
        """
        return self.check_status(task, parent, root, Task.ACTIVE_STATUS)

    def check_complete(self, task=None, parent=None, root=None, status=None):
        """
        Check if a task or a group of tasks are complete
        """
        return self.check_status(task, parent, root, Task.COMPLETE_STATUS)


    def wait(self, task=None, parent=None, root=None, period=None, timeout=None, status=Task.COMPLETE_STATUS):
        """
        Wait while the task or the children or the descendants are active
        :param task: the task to check (id or Task)
        :param parent: the parent of the tasks to check (id or Task)
        :param root: the root of the tasks to check (id or Task)
        :param period: the polling period
        :param timeout: the max time to wait
        :param status: the status or list of status to wait for
        :return: True if all the tasks have finished
        """
        if period is None:
            period = 2

        start_time = time.time()

        while True:
            stati = self.get_all_status(task=task, parent=parent, root=root)
            if all(Task.status_equal(s, status) for s in stati):
                return True
            elif all(Task.status_equal(s, Task.COMPLETE_STATUS) for s in stati):
                # the goal is not matched no more task is active
                return False

            waited_time = int(time.time() - start_time)
            if timeout and waited_time > timeout:
                return False

            self._logger.debug('Wait for tasks (waited_time = {0})'.format(waited_time))
            time.sleep(period)


    def get_records(self, task, max_records=100, increment=True):
        """
        Get the list of records for a given tasks.

        :param task:
        :param max_records:
        :param increment:
        :return:
        """

        if increment and task._last_record_step:
            records = self._torii.graph.get_records(task=task,
                                                    min_step=task._last_record_step + 1,
                                                    max_records=max_records,
                                                    get_last=False)

        else:
            records = self._torii.graph.get_records(task=task,
                                                    min_step=0,
                                                    max_records=max_records,
                                                    get_last=False)

        if records:
            task._last_record_step = int(records[-1].step)

        return records


    def get_viewer_url(self, task):
        """
        Get the URL of the standalone task viewer
        :param task:
        :return:
        """
        # url = '{0}/js/desktop/main-desktop.html?start_tool=taskViewer&id={1}&sessioId={3}'\
        #     .format(self._torii.properties['base_url'], task.id, self._torii.session.id)

        url = '{0}/js/desktop/main-desktop.html?start_tool=taskViewer&id={1}&type={2}&sessionId={3}'\
            .format(self._torii.properties['base_url'],
                    task.application.id, 'application', self._torii.session.id)
        return url


    def get_mapper_url(self, container):
        """
        Get the URL of the standalone task viewer
        :param task:
        :return:
        """
        # https://10.140.9.23:8443/torii/js/desktop/main-desktop.html?start_tool=process_mapper_98895ca2&processMap=5b3e0a5e897b9b3b4e5840d4

        url = '{0}/js/desktop/main-desktop.html?start_tool=process_mapper_98895ca2&processMap={1}&sessionId={2}'\
            .format(self._torii.properties['base_url'],
                    container.id,
                    self._torii.session.id)
        return url


    def open_viewer(self, task):
        """
        Open the standalone task viewer
        :param task:
        :return:
        """
        webbrowser.open_new(self.get_viewer_url(task))


    def open_container_mapper(self, task):
        """
        Open the standalone mapper on the task container
        :param task:
        :return:
        """

        container = self.get(task.root)
        webbrowser.open_new(self.get_mapper_url(container))

    def apply_profile(self, profile_id, task_id):
        try:
            return self._object_class(self.request_post(path='{0}/loadProfile/{1}'.format(task_id, profile_id)), self)
        except ToriiException:
            raise
        except Exception:
            raise ToriiException('Failed to apply the profile {0} on task {1})'.format(profile_id, task_id))




import validate
import pydoc
import os


api_file_src = '/opt/hpcg/core/bin/libhpcg/validate.py'
api_file_src_convert = os.path.realpath(api_file_src)

src_dir = os.path.dirname(os.path.realpath(__file__))
docs_dir = os.path.join(src_dir, 'docs')

if not os.path.exists(docs_dir):
    os.makedirs(docs_dir)

os.chdir(docs_dir)
package = validate
# pydoc.writedocs(src_dir)
pydoc.writedoc('validate')




import sys
import zipfile

from torii.exception import ToriiException
from torii.data import *
from torii.services.torii_service import ToriiService


class AddonService(ToriiService):

    def __init__(self, torii):
        """
        Create the addon service
        """
        ToriiService.__init__(self, torii=torii, name='addons', object_class=ToriiObject)

    def add(self, path):
        """
        Add an addon to Gateway

        :param path: path to the zip file of the addon
        :return: the addon
        """
        if not zipfile.is_zipfile(path):
            raise ToriiException('Failed to add addon. Bad ZIP file: {0}.'.format(path))
        with open(path, 'rb') as fd:
            try:
                r = self._torii.http_session.request(method='POST',
                                                     headers={'Content-Type': 'application/octet-stream'},
                                                     url=self.get_url('add'),
                                                     stream=True,
                                                     data=fd)
                if r.status_code not in [200, 204]:
                    raise ToriiException('Failed to add addon from {0} (code = {1}) \n{2}'
                                         .format(path, r.status_code, r.reason))
                # content =
                # content = json_lib.loads(content)
                # addon = ToriiObject(json=content, service=self)

                return r.json()

            except ToriiException:
                raise
            except Exception as e:
                raise ToriiException('Failed to add addon from {0} \n{1}'
                                     .format(path, e)).with_traceback(sys.exc_info()[2])

    def _updatezip(self, uuid, path):
        """
        Update an existing addon
        :param uuid: identifier of addon
        :param path: path to the zip file of the addon
        :return: the updated addon
        """
        with open(path, 'rb') as fd:
            try:
                r = self._torii.http_session.request(method='POST',
                                                     headers={'Content-Type': 'application/octet-stream'},
                                                     url=self.get_url('{0}/updateZip'.format(uuid)),
                                                     stream=True,
                                                     data=fd)
                if r.status_code not in [200, 204]:
                    raise ToriiException('Failed to update addon from {0} (code = {1}) \n{2}'
                                         .format(path, r.status_code, r.reason))

                return r.json()

            except ToriiException:
                raise
            except Exception as e:
                raise ToriiException('Failed to update addon from {0} \n{1}'
                                     .format(path, e)).with_traceback(sys.exc_info()[2])

    def update(self, addon, create=False, **kwargs):
        """
        Update an existing addon
        :param addon: the addon
        :param create: False
        :param kwargs: extra parameters (zip=path)
        :return: the updated addon
        """

        # Search fields:
        if kwargs:
            for key, value in kwargs.items():
                # Set field: zip
                if key == 'zip':
                    if not zipfile.is_zipfile(value):
                        raise ToriiException('Failed to update addon. Bad ZIP file: {0}.'.format(value))
                    if isinstance(addon, ToriiObject):
                        uuid = addon.uuid
                        return self._updatezip(uuid=uuid, path=value)
                    else:
                        raise ToriiException('Failed to update addon. Bad JSON addon: {0}.'.format(addon))

        return ToriiService.update(self, addon, create=False)



from torii.exception import ToriiException
from torii.data import *
from torii.services.torii_service import ToriiService


class ProfileService(ToriiService):

    def __init__(self, torii):
        """
        Create the Profile service.

        :param torii:
        """
        ToriiService.__init__(self, torii=torii, name='profiles', object_class=ToriiObject)

    def create_profile(self, obj, task_id=''):
        """
                     Create a profile in Torii.
                     :param obj: the object to register.
                     :param taskId: task id, must be in the db.
                     :return: the modified object.
                     """

        params = {
            'taskId': task_id
        }

        try:
            if isinstance(obj, self._object_class):
                json = self.request_post(json=obj.to_json(), params=params)
                # update current object
                obj.from_json(json)
                return obj
            else:
                json = self.request_post(json=obj, params=params)
                # create a new ToriiObject
                return self._object_class(json, self)

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException(
                'Failed to create a {0} in the system\n{1}'.format(self._object_class, e)).with_traceback(
                sys.exc_info()[2])

    def update(self, obj, create=False, **kwargs):
        """
        Update an object
        :param obj:
        :return:
        """

        if create:
            try:
                self.get(obj.id)
            except:
                return self.create(obj, **kwargs)

        try:
            if isinstance(obj, self._object_class):
                if obj.has_changed() or 'taskId' in kwargs:
                    # update the remote version
                    json = self.request_post(path=obj.id, json=obj.to_json(), params=kwargs)
                else:
                    # update local version
                    json = self.request_get(path=obj.id)
            else:
                raise ToriiException('unexpected case: please think about it before removing tis exception :-)')
                json = self.request_post(path=obj['_id'], json=obj, params=kwargs)

            obj.from_json(json)

            return obj

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to update {0}\n{1}'.format(obj, e)).with_traceback(sys.exc_info()[2])

    def _prepare_params(self, params):
        prepared = {}
        if params is not None:
            for name, value in params.items():
                if value is True:
                    value = 'true'
                elif value is False:
                    value = 'false'
                prepared[name] = value
        return prepared



from datetime import datetime as dt

from bson import ObjectId
from pymongo import ASCENDING, DESCENDING


class AbstractDao:
    """
    This class offers the base MongoDB data access functionality.

    Any custom query function should use the helper methods to guarantee functionality.
    """

    PROHIBITED_TO_UPDATE = ['_id', 'location', 'creationDate', 'canEdit', 'canRemove']

    @staticmethod
    def get_collection(db_torii, db_bo, collection_name):
        """
        Get the correct BO database collection from the bo_classes collection.

        :param db_torii: Torii database object
        :param db_bo: Customer database object
        :param collection_name: BO Class name
        :return: Mongo collection from db_bo param
        """
        collection = db_torii.get_collection('bo_classes')
        return db_bo.get_collection(collection.find_one({'name': collection_name})['collection'])

    def find_one(self, collection, query: dict, include=[]) -> dict:
        """
        Find one document in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param include: List of keys to project in final result (removes all other keys from result).
        :return: The document if found, otherwise None.
        """
        query = self.prepare_query(query)

        return self.prepare_result(collection.find_one(query, projection=include))

    def find_many(self, collection, query,
                  filters={}, include=None, direction='ASC', sort=None, timeout=300, limit=0, start=0) -> list[dict]:
        """
        Find one or more documents in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param filters: Optional personalised filters.
        :param include: List of keys to project in final result (removes all other keys from result).
        :param direction: Orders the result as Ascending ('ASC') or Descending (any other value).
            Only applied if 'sort' is not None. Default 'ASC'.
        :param sort: Key to sort the result. Applies 'direction'.
        :param timeout: Time limit (in seconds) for the query operation to return a result.
        :param limit: Limit the number of documents returned by the query.
        :param start: The number of results to skip, therefore starting at a later part of the result.
        :return: List of dictionary documents.
        """
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))

        cursor = collection.find(query, projection=include)
        if sort: cursor.sort(sort, ASCENDING if direction == 'ASC' else DESCENDING)
        cursor.skip(start)
        cursor.limit(limit)

        if timeout > 0: cursor.max_time_ms(timeout * 1000)

        return [self.prepare_result(document) for document in cursor]

    def count_documents(self, collection, filters, start=0, limit=0) -> int:
        """
        Count the number of documents returned by query.

        :param collection: Collection context.
        :param filters: Optional personalised filters.
        :param start: The number of results to skip, therefore starting at a later part of the result.
        :param limit: Limit the number of documents returned by the query.
        :return: Total number of documents found by query.
        """
        query = self.prepare_filter(filters)

        kwargs = {'filter': query}
        if start: kwargs['skip'] = start
        if limit: kwargs['limit'] = limit

        return collection.count_documents(**kwargs)

    def insert_one(self, collection, document: dict) -> dict:
        """
        Insert one document into the collection.

        :param collection: Collection context.
        :param document: JSON like document.
        :return: The given document.
        """
        doc = self.prepare_query(document)

        collection.insert_one(doc)

        return document

    def insert_many(self, collection, documents: list[dict]):
        """

        :param collection:
        :param documents:
        :return:
        """
        doc = [self.prepare_query(document) for document in documents]

        collection.insert_many(doc)

    def update_and_retrieve_one(self, collection, query: dict, document: dict, filters={}):
        """
        Update one document in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param document: JSON like document.
        :param filters: Optional personalised filters.
        :param retrieve: Retrieve updated document.
        :return: The document after the update if found, otherwise None.
        """
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))
        document = self.prepare_update(document)

        return self.prepare_result(
                collection.find_one_and_update(filter=query, update={'$set': document}, return_document=True))

    def update_one(self, collection, query: dict, document: dict, filters={}):
        """
        Update one document in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param document: JSON like document.
        :param filters: Optional personalised filters.
        :param retrieve: Retrieve updated document.
        :return: The document after the update if found, otherwise None.
        """
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))
        document = self.prepare_update(document)

        collection.update_one(filter=query, update={'$set': document})

    def delete_one(self, collection, query: dict, filters={}) -> None:
        """
        Delete one document in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param filters: Optional personalised filters.
        :return: None
        """
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))

        collection.delete_one(query)

    def delete_many(self, collection, query: dict, filters={}) -> None:
        """
        Delete one or more documents in collection.

        :param collection: Collection context.
        :param query: Query filter.
        :param filters: Optional personalised filters.
        :return: None
        """
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))

        collection.delete_many(query)

    @staticmethod
    def prepare_query(query: dict) -> dict:
        """
        Helper method.

        Guarantees the ID in the query/document is a MongoDB ObjectId.

        :param query: Query filter or JSON like document.
        :return: The modified query/document.
        """
        if '_id' in query and isinstance(query['_id'], str):
            query['_id'] = ObjectId(query['_id'])

        return query

    @staticmethod
    def prepare_filter(filters) -> dict:
        """
        Helper method.

        Turns the optional, personalised filters into a query compatible dictionary.

        :param filters: Optional personalised filters.
        :return: A query compatible dictionary.
        """
        query = {}

        if len(filters) == 1:
            item = filters[0]
            query = {
                item['property']:
                    item['value'] if item['operator'] == 'eq'
                    else {'$regex': item['value']} if item['operator'] == 'like'
                    else {f'${item["operator"]}': item['value']}
            }
        elif len(filters) > 1:
            query = {"$and": [
                {
                    item['property']:
                        item['value'] if item['operator'] == 'eq'
                        else {'$regex': item['value']} if item['operator'] == 'like'
                        else {f'${item["operator"]}': item['value']}
                }
                for item in filters
            ]}

        return query

    def prepare_update(self, document: dict):
        """
        Helper method.

        Removes any key in the given document that is prohibited from being updated by any User.
        Also adds/updates the 'modificationDate' value to reflect the new update.

        :param document: JSON like document.
        :return: The modified document.
        """
        [document.pop(key) for key in self.PROHIBITED_TO_UPDATE if key in document]

        document['modificationDate'] = int(dt.now().timestamp())

        return document

    @staticmethod
    def prepare_result(document):
        """
        Helper method.

        Prepares a MongoDB output result to be JSON compatible.

        :param document: JSON like document.
        :return: The modified document.
        """
        if not document:
            return document
        document['_id'] = str(document['_id'])
        return document

    @staticmethod
    def to_torii_object(document: dict):
        """
        Helper method.

        Prepares a document by converting all keys to String
        and adding all default elements expected from a ToriiObject.

        :param document: JSON like document.
        :return: The modified document with all default elements expected from a ToriiObject.
        """

        def convert_key_to_string(doc: dict):
            return {str(k): convert_key_to_string(i) if isinstance(i, dict) else i for k, i in doc.items()}

        doc = convert_key_to_string(document)

        return {
            'location': doc.pop('location') if 'location' in doc else 'main',
            'name': doc.pop('name') if 'name' in doc else '',
            'creationDate': doc.pop('creationDate') if 'creationDate' in doc else int(dt.now().timestamp()),
            'modificationDate': doc.pop('modificationDate') if 'modificationDate' in doc else int(dt.now().timestamp()),
            'description': doc.pop('description') if 'description' in doc else '',
            'wikiURI': doc.pop('wikiURI') if 'wikiURI' in doc else '',
            'tags': doc.pop('tags') if 'tags' in doc else [],
            'users': doc.pop('users') if 'users' in doc else [],
            'teams': doc.pop('teams') if 'teams' in doc else [],
            'projects': doc.pop('projects') if 'projects' in doc else [],
            'createdBy': doc.pop('createdBy') if 'createdBy' in doc else None,
            'content': doc
        }



import sys
from base64 import b64encode
from json import dumps
from secrets import token_bytes

import requests
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from torii.exception import ToriiException
from torii.services.torii_service import ToriiService
from torii.data.torii_object import ToriiObject


class SessionService(ToriiService):

    def __init__(self, torii):
        """
        Create the session service
        """
        ToriiService.__init__(self, torii=torii, name='sessions', object_class=ToriiObject)


    def check_auth(self, username=None, password=None):
        credentials = SessionService._build_credentials(username, password)

        try:
            json = self.request('POST', path='login', json=credentials)
        except Exception as e:
            raise ToriiException('Failed to log as {0} on {1} : {2}'
                               .format(username, self._torii.properties['base_url'], e)).with_traceback(sys.exc_info()[2])

        session = ToriiObject(json, self)
        self._logger.info('User logged as {0} on {1}'.format(session.user.name, self._torii.properties['base_url']))

        return session


    def login(self, username=None, password=None, timeout=30, oauth=False):
        credentials = SessionService._build_credentials(username, password)

        try:
            if oauth:
                credentials = SessionService._encrypt_credentials(credentials)
                json = self.request_post(path='microsoftApiLogin', json=credentials, timeout=timeout)
            else:
                json = self.request_post(path='login', json=credentials, timeout=timeout)
        except Exception as e:
            raise ToriiException('Failed to log as {0} on {1} : {2}'
                                 .format(username, self._torii.properties['base_url'], e)).with_traceback(sys.exc_info()[2])

        session = ToriiObject(json, self)

        self._logger.info('User logged as {0} on {1}'.format(session.user.name, self._torii.properties['base_url']))

        return session

    @staticmethod
    def _encrypt_credentials(creds: dict):
        cred_str = dumps(creds).encode('utf-8')
        key = AESGCM.generate_key(bit_length=256)  # Encryption Key
        iv = token_bytes(12)  # Initialization Vector
        encrypted_data = AESGCM(key).encrypt(iv, cred_str, None)  # Encrypted data

        payload = {
            'a': b64encode(encrypted_data).decode('utf-8'),
            'b': b64encode(key).decode('utf-8'),
            'c': b64encode(iv).decode('utf-8')
        }

        return payload

    def su(self, user):
        if isinstance(user, ToriiObject):
            params = {'userId': user.id}
        else:
            params = {'userId': user}

        try:
            json = self.request_post(path='su', params=params)
        except Exception as e:
            raise ToriiException('Failed to switch user with {} : {}'
                               .format(user, e)).with_traceback(sys.exc_info()[2])

        session = ToriiObject(json, self)

        self._logger.info('User logged as {0} on {1}'.format(session.user.name, self._torii.properties['base_url']))
        self._torii.session = session

        if self._torii.properties['use_python_dao']:
            self._torii.user_manager.update_user()

        return session


    def logout(self):
        try:
            self.request_post(path='logout')
        except Exception as e:
            raise ToriiException('Failed to logout : {}'
                               .format(self._torii.properties['base_url'], e))\
                .with_traceback(sys.exc_info()[2])

        self._logger.info('User logged out from {}'.format(self._torii.properties['base_url']))


    def check(self, session_id=None, additional_cookies={}):
        if session_id:
            params = {'ToriiSessionId': session_id}
        else:
            params = None

        try:
            url = self.get_url('check')
            http_session = requests.Session()
            http_session.trust_env = False
            http_session.verify = False
            http_session.proxies.update({})
            http_session.cookies.update({'ToriiSessionId': session_id, **additional_cookies})
            response =   http_session.get(url=url, params=params)
            assert response.status_code == 200
            json = response.json()
            json = self.request_get(path='check', params=params)
        except Exception as e:
            self._logger.warn('No active session {} on {}'.format(session_id, self._torii.properties['base_url']))
            return None

        session = ToriiObject(json, self)
        self._logger.info('User {0} has an active session on {1} (id={2})'
                          .format(session.user.name, self._torii.properties['base_url'], session.id))

        return session

    @staticmethod
    def _build_credentials(username: str, password: str):
        username = username.strip()
        password = password.strip()

        return {
            'userName': username,
            'password': password
        }



import getpass

import bson
import copy
import inspect
import logging
import os
import requests
from pymongo import MongoClient
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import json

import sys

import time

from torii.managers.dao_manager import DaoManager
from torii.managers.user_manager import UserManager
from torii.services import *
from torii.data import *
from torii.exception import *
from torii.services.bs_service import BSService
from torii.session import CustomSession


def _mkdir_recursive(path):
    sub_path = os.path.dirname(path)
    if not os.path.exists(sub_path):
        _mkdir_recursive(sub_path)
    if not os.path.exists(path):
        os.mkdir(path, 0o700)

class Torii(object):

    # the list of services proposed by the torii session
    # the value may correspond to :
    #  - a class: the service will be instanciated with the class constructor
    #  - a path: the service will be instanciated with the default service constructor using this path
    #  - nothing: the service will be instanciated with the default service constructor and default values
    __GENERIC_SERVICES = {
        'tasks': TaskService,
        'sessions': SessionService,
        'addons': AddonService,
        'applications': ApplicationService,
        'users': None,
        'bo_classes': None,
        'proxies': None,
        'scripts': None,
        'servers': None,
        'clusters': None,
        'teams': None,
        'projects': None,
        'mountpoints': None,
        'picoms': PicomService,
        'profiles': ProfileService,
        'transfers':None,
        'snapshots': None
    }

    __GENERIC_ADDON_SERVICES = {
        '98895ca2' : {'name': 'maps', 'service': 'process_mapper_98895ca2'},
        'c9a0d5e4': {'name': 'ai', 'service': AiService}
    }

    # the list of properties and their default
    __DEFAULT_PROPERTIES = {
        'base_url': 'https://localhost:8443/torii',
        'current_container': None,
        'cluster': None,
        'username': os.environ.get('USER'),
        'session_id': None,
        'private_key': os.path.expanduser('~/.ssh/id_rsa'),
        'mongo_connection_string': None,
        'use_python_dao': False,
        'use_oauth_login': False
    }

    @property
    def logger(self):
        return self._logger

    @property
    def base_url(self):
        return self.properties['base_url'] if 'base_url' in self.properties else None

    def __init__(self, properties_file=None, login=None, timeout=30,
                 username=None, password=None,
                 container=None, name=None,
                 session_id=None,
                 log_level=logging.DEBUG, save_properties=None, prompt_password=True,
                 logger=None, http_cookies=None, http_timeout=3, http_max_retries=5, db_prefix=None,
                 **kwargs):
        """

        :param properties_file:  path to the property file to use
        :param username:         username (optional)
        :param password:         password (optional)
        :param session_id        initiate the client with an existing session
        :param login:            autologin
                                 (True=force the logging, False=no logging, None=try to reuse the existing session or log in)
        :param timeout:          autologin timeout
        :param container:        use container ('new', 'current' or None)
        :param name:             name of the session
        :param log_level:        'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'
        :param save_properties:  save the connection settings in the properties file
                                 (True=force the writing, False=no writing, None=only the session)
        :param http_cookies:     Cookies dict to be appended to HTTP Session
        :param http_timeout:     Default timeout for all HTTP Session requests
        :param http_max_retries: Maximum number of retries for the HTTP Session requests
        :param db_prefix:        Customer database name to filter for in bo_bases
        :param kwargs:           extra connection properties (see __DEFAULT_PROPERTIES)
        """

        if logger:
            self._logger = logger
        else:
            self._logger = logging.getLogger('torii')

            # init a default handler for the logs
            self._logger.setLevel(log_level)
            ch = logging.StreamHandler()
            ch.setLevel(log_level)

            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)

            self._logger.addHandler(ch)

        self.init_properties(username, password, properties_file, **kwargs)

        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        self.http_session = CustomSession()
        self.http_session.trust_env = False
        self.http_session.verify = False
        self.http_session.proxies.update({})
        self.http_session.cookies.update(http_cookies or {})

        self.session = None
        self.current_map = None
        self.current_container = None
        self.ignore_save_properties = save_properties is not None and not save_properties
        self.save_only_session_properties = save_properties is None

        # initialize database access
        self.mongo_client = MongoClient(self.properties['mongo_connection_string'])
        self.mongo_database_torii = self.mongo_client.get_database('Torii')
        bo = self.mongo_database_torii.get_collection('bo_bases').find_one(filter={'name': db_prefix})
        self.mongo_database_torii_bo = self.mongo_client.get_database(bo['basename'] if bo else 'Torii_BO')

        self.user_manager = UserManager(self)
        self.dao_manager = DaoManager(self)

        # initialize specific services
        self.users =  ToriiService(self, 'users')
        self.proxies =  ToriiService(self, 'proxies')
        self.profiles =  ToriiService(self, 'profiles')
        self.scripts =  ToriiService(self, 'scripts')
        self.servers =  ToriiService(self, 'servers')
        self.bo_classes =  ToriiService(self, 'bo_classes')
        self.bo_bases =  ToriiService(self, 'bo_bases')
        self.clusters =  ToriiService(self, 'clusters', object_class=Cluster)
        self.teams =  ToriiService(self, 'teams')
        self.projects =  ToriiService(self, 'projects')
        self.mountpoints =  ToriiService(self, 'mountpoints')
        self.addons =  ToriiService(self, 'addons')
        self.transfers = ToriiService(self, 'transfers')
        self.organisations = OrganisationService(self)
        self.snapshots =  ToriiService(self, 'snapshots')
        self.shortcuts =  ToriiService(self, 'shortcuts')
        self.addons = ToriiService(self, 'addons')

        # initialize specific services
        self.business_services = BSService(self)
        self.files = FileService(self)
        self.graph = GraphService(self)
        self.tasks =  TaskService(self)
        self.sessions =  SessionService(self)
        self.applications =  ApplicationService(self)
        self.picoms =  PicomService(self)
        self.profiles =  ProfileService(self)
        self.addons = AddonService(self)

        # try to reuse former session
        if session_id is not None or login is None:
            self.update_session(session_id=session_id)

        # try to log automatically
        if not self.session and (login is None or login):
            self.login(prompt_password=prompt_password, timeout=timeout)
            if not self.session:
                raise ToriiException('Failed to auto-log in, please check the configuration file {0}'.format(self.property_file))

        if self.session:
            self.save_properties()

            # prepare the current container
            if container is 'current' and 'current_container' in self.properties and self.properties['current_container']:
                # reuse the former container
                self.current_container = self.tasks.get(self.properties['current_container'])
            elif container in ['new', 'current'] :
                # the user explicitely asked for a new coontainer
                self.new_container(name=name, cluster=self.properties['cluster'])
            elif container is not None:
                raise ToriiException('Unexpected container argument \'{0}\''.format(container))

        if self.properties['use_python_dao']:
            self._logger.info(f"Using Direct DAO (use_python_dao={self.properties['use_python_dao']})")
            self.user_manager.update_user()


    def init_properties(self, username=None, password=None, property_file=None, **kwargs):
        """
        Init the properties from the property file
        :param property_file:
        :return:
        """
        self.properties = copy.deepcopy(self.__DEFAULT_PROPERTIES)

        # get the propety file path
        if property_file:
            self.property_file = property_file
        elif 'GATEWAY_PROPERTIES' in os.environ:
            self.property_file = os.environ['GATEWAY_PROPERTIES']
        elif username:
            self.property_file = os.path.expanduser('~/.torii/{}.properties'.format(username))
        else :
            self.property_file = os.path.expanduser('~/.torii/default.properties')

        # add password and username to the properties
        if username:
            self.properties['username'] = username
        if password:
            self.properties['password'] = password

        # read the property file if it exists
        if os.path.isfile(self.property_file):
            try:
                self._logger.info('Loading properties from {}'.format(self.property_file))
                properties = json.load(open(self.property_file, 'r'))

                for key, value in properties.items():
                    if key in self.__DEFAULT_PROPERTIES:
                        self.properties[key] = value
                    else:
                        self._logger.warning('Unknown property {0} = {1}'.format(key, value))

            except Exception as e:
                self._logger.warning('Failed to read the property file {0}\n{1}'.format(self.property_file, e))

        # force properties from the arguments
        for key, value in kwargs.items():
            if (key in self.__DEFAULT_PROPERTIES or key == 'password') and value is not None:
                self.properties[key] = value
            else:
                self._logger.warning('Unknown property {0} = {1}'.format(key, value))


    def get_bo_service(self, bo_name, object_class=BusinessObject):
        """
        Get a BO service.

        :param bo_name: the name of the Business Object class
        :return:
        """
        return BusinessObjectService(self, bo_name=bo_name, object_class=object_class)

    def list_bo_classes(self, explore=False):
        """
        List the BO services available

        :return:
        """
        root_bo_service = Service(self, 'bo')
        bo_class_names = root_bo_service.request_get(params={'explore':explore})

        return bo_class_names

    def save_properties(self, only_new=False):
        """
        Save the property file
        """
        try:
            # check if the property file must be overwriten at all
            if self.ignore_save_properties:
                self._logger.info('The properties wont be saved to {} (option save_properties=False)'.format(self.property_file))
                return

            # check if the property file already exists
            file_exists = os.path.isfile(self.property_file)
            if file_exists and self.save_only_session_properties:
                if self.session:
                    self.update_saved_properties('session_id', self.session.id)
                return


            self._logger.info('Saving properties to {}'.format(self.property_file))

            # prepare the properties to save
            saved_properties = dict(self.properties)
            if 'password' in saved_properties:
                del saved_properties['password']

            # create the directory if it does not exist
            _mkdir_recursive(os.path.dirname(self.property_file))
            with open(self.property_file, 'w+') as f:
                json.dump(saved_properties, f, indent=4, sort_keys=True)
                f.write('\n')

            os.chmod(self.property_file, 0o600)

        except Exception as e:
            raise ToriiException('Failed to write the property file {0}\n{1}'.format(self.property_file, e))\
                .with_traceback(sys.exc_info()[2])

    def update_saved_properties(self, key, value):

        try:
            self.properties[key] = value

            # check if the property file already exists
            file_exists = os.path.isfile(self.property_file)

            if not file_exists:
                return self.save_properties()

            self._logger.info('Updating properties to {}'.format(self.property_file))

            # read the saved property file
            with open(self.property_file, 'r') as f:
                saved_properties = json.load(f)

            # update the property file
            saved_properties[key] = value
            with open(self.property_file, 'w+') as f:
                json.dump(saved_properties, f, indent=4, sort_keys=True)
                f.write('\n')

        except Exception as e:
            raise ToriiException('Failed to update the property file {0}\n{1}'.format(self.property_file, e))\
                .with_traceback(sys.exc_info()[2])

    def update_session(self, session_id = None):
        self.session = None

        if session_id:
            # we explicitely try to udate the session so we must excet on error
            except_on_fail = True
        elif 'session_id' in self.properties:
            # in this case it is best effort
            session_id = self.properties['session_id']
            except_on_fail = False
        else:
            return None

        try:
            self.session = self.sessions.check(session_id, self.http_session.cookies)
        except Exception as e:
            if  except_on_fail:
                raise   Exception(f'Failed to reuse session {session_id}') from e
            else:
                self._logger.info(f'Failed to reuse session {session_id}, try to initiate a new session')

        if self.session:
            self.properties['session_id'] = self.session.id
            self._logger.info('Re-use session on {} as {} (created on {})'
                              .format(self.properties['base_url'],
                                      self.session.user.name,
                                      time.ctime(self.session.creationDate//1000).strip()))

        return self.session

    def login(self, username=None, password=None, prompt_password=False, retry_prompt=None, timeout=30, retry=2):
        # clear the session
        self.session = None

        login=''
        if username:
            login = username
        elif 'username' in self.properties and self.properties['username']:
            login = self.properties['username']

        secret = None

        # initialize the number of prompt retries
        if prompt_password and retry_prompt is None :
            retry_prompt = 3

        if password:
            secret = password
            comment = 'password = XXXXX'
        elif 'password' in self.properties and self.properties['password']:
            secret = self.properties['password']
            comment = 'password = XXXXX'
        elif 'private_key' in self.properties \
                and self.properties['private_key'] \
                and os.path.isfile(self.properties['private_key']):
            secret = open(self.properties['private_key'], 'r').read().rstrip('\n')
            # comment = 'private_key = {}'.format(self.properties['private_key'])
            comment = 'password = XXXXX'

        if secret is None:
            if not retry_prompt :
                self._logger.info('No more retry')
                return None
            elif not prompt_password:
                return None
            else:
                secret = getpass.getpass('Password (login={}): '.format(login))
                # comment = 'password = {}'.format(secret)
                comment = 'password = XXXXX'

        self._logger.info('Login on {} as {} ({})'
                          .format(self.properties['base_url'], login, comment))

        # initialize all the known torii services
        #for name, service in self.__GENERIC_SERVICES.items():
        #    self._expose_service(name=name, service=service)
        try:
            self.session = self.sessions.login(login, secret, timeout=timeout, oauth=self.properties['use_oauth_login'])
        except Exception as e:
            self._logger.error(str(e).splitlines()[0])

        if self.session:
            self.properties['session_id'] = self.session.id

            # initialize all the known torii addons services
            for addon in self.addons.list():
                if addon.installed and addon.uuid in self.__GENERIC_ADDON_SERVICES:
                    self._expose_service(**self.__GENERIC_ADDON_SERVICES[addon.uuid])

        elif prompt_password and retry_prompt is not None  and retry_prompt > 0:
            # retry to log with prompting
            self.login(username=username, retry_prompt=retry_prompt-1, prompt_password=prompt_password, timeout=timeout)
        elif retry is not None and retry > 0:
            # retry to log
            self.login(username=username, password=password, prompt_password=False, timeout=timeout, retry=retry-1)

        return self.session

    def _expose_service(self, name, service):
        """
        Expose a service as a Torii attibute
        :param name: the attribute name
        :param service: the service class or the path
        :return:
        """
        if inspect.isclass(service):
            setattr(self, name, service(self))
        else:
            setattr(self, name, ToriiService(self, name, service))

    def service(self, name, path=None):
        return ToriiService(self, name, path)


    def set_current_container(self, container=None, num=None):
        if isinstance(container, ToriiObject):
            self.current_container = container
        elif isinstance(container, str):
            self.current_container = self.tasks.get(id=container)
        elif num:
            self.current_container = self.tasks.get(num=num)
        else:
            raise ToriiException('Missing parameters')

        self.properties['current_container'] = self.current_container.id
        self.save_properties()

        return self.current_container


    def new_container(self, name=None, map=True, cluster=None):
        """
        Initiate a new container to use as current container
        TODO we should create the container independently of the
        """
        if name is None:
            name = 'Session for {0} ({1})'\
                .format(self.session.user.name, time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

        self.current_container = self.tasks.new(type='container', name=name)


        # try to get the default cluster
        cluster_obj = None
        if isinstance(cluster, ToriiObject):
            cluster_obj = self.current_container.set_cluster(cluster)
        elif isinstance(cluster, str):
            if bson.objectid.ObjectId.is_valid(cluster):
                cluster_obj = self.clusters.get(cluster)
            else:
                cluster_obj = self.clusters.get(name=cluster)
        else:
            try:
                cluster_obj = self.clusters.get()
            except:
                pass

        if cluster_obj:
            self.current_container.set_cluster(cluster_obj)

        self.properties['current_container'] = self.current_container.id
        self.save_properties()

        return self.current_container


import inspect

import re

from torii.data.torii_object import ToriiObject

def fun2script(function):

    imports = ''

    lines = inspect.getsourcelines(function)
    index = 0
    for line in lines[0]:
        if index == 0:
            # parse the function header
            m = re.search('(^ *)def (.*)\((.*)\) *:', line)
            indent = '^' + m.group(1)
            function_name = m.group(2)
            function_args = m.group(3)

            # rewrite the function header without indent
            function_def = 'def {0}({1}):\n'.format(function_name, function_args)
        else:
            # remove the indents
            line = re.sub(indent, '', line)

            # check if the line is an import
            if re.match(' *import .*', line) or re.match(' *from .* import .*', line):
                imports += '{0}\n'.format(line.strip())

            else:
                function_def += line

        index += 1


    # copy the import we found in the function
    script = '{0}\n'.format(imports)

    # add missing import
    if 'PolicyExecutor' not in imports:
        script += '''
from torii.executor import PolicyExecutor
'''

    # add the function definition
    script += '\n{0}\n'.format(function_def)

    # add the main part
    script += '''
if __name__ == "__main__":

'''
    # we consider the first argument of the function as the executor
    # without arg, we assume the task is pasrsed in the funstion
    task = None
    if function_args is not None and function_args.strip():
        args = function_args.split(',')
        task = args[0]
        script += '''
    {0} = PolicyExecutor()
    
    '''.format(task)


    # call the function
    script += '''
    {0}({1})
    '''.format(function_name, task if task else '')

    return script



class Cluster(ToriiObject):

    def __init__(self, json, service=None, details=False):
        ToriiObject.__init__(self, json, service)

    def get_sched_option(self, name):
        """ Get the sched_option object based on its name """
        # get the option from the task
        option = next((option for option in self.scheduler.options
                       if option.label.lower().strip() == name.lower().strip()), None)

        return option

    def get_sched_option_names(self):
        """ Get the sched_option object based on its name """
        return [option.label.lower().strip() for option in self.scheduler.options]


    def set_policy(self, script):
        if callable(script):
            self.policyScript = fun2script(script)
        else:
            self.policyScript = script


from torii.data.record import Record
from torii.services import Service


class GraphService(Service):

    def __init__(self, torii):
        """
        Create the graph service
        """
        Service.__init__(self, torii=torii, name='graph')


    def get_records(self, task, min_step=0, max_records=100, get_last=False):
        """
        Get the list of records for a given tasks.

        :param task:        The task
        :param min_step:    The minimum value of the step
        :param max_records: The max number of records to get
        :param get_last:    Give the last records in priority

        :return: a list of records
        """

        params = {
            'stepMin': min_step,
            'nbMax': max_records,
            'last': get_last
        }

        json = self.request_get(path=task.id, params=params)

        return [Record(x) for x in json['taskRecords']]


import copy
import sys

from collections.abc import Iterable

from torii.exception import ToriiException
from torii.data.torii_object import ToriiObject, Struct


class Task(ToriiObject):

    STATUS = ['REGISTERED', 'RUNNING', 'PAUSED', 'WAITING', 'SUBMITTING', 'SCHEDULING', 'UNKNOWN', 'FINISHED', 'FAILED', 'CANCELLED']
    ACTIVE_STATUS = ['REGISTERED', 'RUNNING', 'PAUSED', 'WAITING', 'SUBMITTING', 'SCHEDULING', 'UNKNOWN']
    COMPLETE_STATUS = ['FINISHED', 'FAILED', 'CANCELLED']
    FAILURE_STATUS = ['FAILED', 'CANCELLED']

    def __init__(self, json={}, service=None):
        ToriiObject.__init__(self, json, service)

        self._cluster_obj = None
        self._last_record_step = None

    def __str__(self):
        """ String converter """
        if self.num:
            return 'task {0}'.format(self.num)
        else:
            return 'new task'

    @property
    def step(self):
        if hasattr(self, 'runInfos') and hasattr(self.runInfos, 'step'):
            return self.runInfos.step
        else:
            return None

    @classmethod
    def status_equal(cls, status, ref_status):
        if not status or not ref_status:
            return False
        elif isinstance(ref_status, Iterable):
            return status in ref_status
        else:
            return status == ref_status
        pass

    def is_status(self, status, refresh=False):

        """
        Tells if the task status correspond to a given status or list of status
        :param refresh:

        :return True if the task is active
        """
        if refresh:
            self.refresh()

        return Task.status_equal(self.status, status)

    def is_active(self, refresh = False):
        return self.is_status(Task.ACTIVE_STATUS, refresh=refresh)

    def is_complete(self, refresh = False):
        return self.is_status(Task.COMPLETE_STATUS, refresh=refresh)

    def wait(self, period=None, timeout=None, status=COMPLETE_STATUS):
        return self._service.wait(task=self, period=period, timeout=timeout, status=status)

    def _get_input(self, name):
        """ Get the input object based on its name """
        try:
            input = next(input for input in self.inputs if input.name == name)
        except:
            raise ToriiException('Cannot find input {0} in {1}'.format(name, self)).with_traceback(sys.exc_info()[2])

        return input

    def _get_output(self, name):
        """ Get the output object based on its name """
        try:
            output = next(output for output in self.outputs if output.name == name)
        except:
            raise ToriiException('Cannot find output {0} in {1}'.format(name, self)).with_traceback(sys.exc_info()[2])

        return output

    def clear_input(self, name):
        """
        Clear the value or the file list of an input

        :param name: the name of the input
        :return:
        """
        input = self._get_input(name)
        delattr(input, 'value')
        delattr(input, 'files')

    def set_input(self, name, value=None, path=None, server=None, append=True, link=False):
        """
        Set an input value based on its name

        :param name: the name of the input
        :param value: the value to set
        :param path:
        :param server:
        """
        input = self._get_input(name)

        if input.type == 'file':
            if isinstance(value, list):
                setattr(input, 'files', value)
            else:
                if isinstance(value, Struct):
                    file = value
                elif path:
                    # prepare the file description based on path and server
                    file = Struct({
                        'action': 'LINK' if link else 'COPY',
                        'path': path,
                        'server': server.ref if isinstance(server, ToriiObject) else server
                    })
                else:
                    raise ToriiException('Cannot set input \'{0}\' ad payload'.format(name))


                # append or reset the file list
                if append and hasattr(input, 'files') and input.files is not None:
                    input.files.append(file)
                else:
                    setattr(input, 'files', [file])

        elif input.type != "file" and value is not None:
            # careful value attribute may not exist
            setattr(input, 'value', value)
        else:
            raise ToriiException("Parameters does not match the input type")

    def get_input(self, name):
        """
        Get an input value based on its name

        :param name: the name of the input 
        :return: the value
        """
        input = self._get_input(name)\

        if input.type == 'file' and hasattr(input, 'files'):
            return input.files
        elif hasattr(input, 'value'):
            return input.value
        else:
            return None

    def map_input(self, name, task=None, input=None, output=None, mapping=None):
        """
        Map an input value to another task input of output
        :param name:    the input to map
        :param task:    the previous task to map from
        :param input:   the input of the previous task
        :param output:  the output of the previous task
        :param mapping: the raw mapping formula (for complex mapping)
        :return:
        """
        from_input = self._get_input(name)

        if isinstance(task, Task) or isinstance(task, str):
            id = task if isinstance(task, str) else task.id
            if input:
                task._get_input(input)
                from_input.mapping = id + ':' + input
            elif output:
                #task._get_output(output)
                from_input.mapping = id + ':' + output
            else:
                raise ToriiException('Bad payload, missing input/output name')
        elif mapping:
            from_input.mapping = mapping
        else:
            raise ToriiException('Bad payload, missing task + input/output name or a mapping formula')


    def clear_dependencies(self):
        """
        Clear the list of dependencies of the task
        """
        self.conditions = []

    def get_dependencies(self):
        """
        Get the list of dependencies of the task
        """
        return [condition for condition in self.conditions if condition.type == 'dependency']

    def add_dependency(self, task=None, criterion='AFTER_SUCCESS', orMode=False, keepEnvironment=True):
        """
        Add a dependency on one or several tasks
        :param task: the task we depend on
        :param criterion: the validity criterion
        :param orMode: is the dependency necessary or sufficient
        :param keepEnvironment: does the task inherit the context from this dependency
        """

        # we can add several dependencies at once
        if isinstance(task, list):
            for t in task :
                self.add_dependency(t, criterion, orMode, keepEnvironment)

            return
        elif isinstance(task, ToriiObject):
            task = task.ref

        valid_criteria = ['AFTER_ANY', 'AFTER_SUCCESS', 'AFTER_FAILED', 'IN_PARALLEL']
        if criterion not in valid_criteria:
            raise ToriiException('Bad parameter, the criterion must be in : {0}'.format(" ".join(valid_criteria)))
        elif not isinstance(task, Struct):
            raise ToriiException('Bad parameter, the task is invalid')

        condition = Struct({
            'type': 'dependency',
            'taskReference': task,
            'criterion': criterion,
            'keepEnvironment': keepEnvironment,
            'orMode': orMode
        })
        self.conditions.append(condition)

    def get_output(self, name):
        """
        Get an output value based on its name

        :param name: the name of the output 
        :return: the value
        """
        return self._get_output(name).value

    def set_context(self, name, value):
        """
        Set a context value

        :param name: the context property
        :param value: the value to set
        """

        # careful value attribute may not exist
        setattr(self.context, name, value)

    @property
    def cluster_obj(self):
        """
        Get the cache cluster object
        :return:
        """
        if self.cluster is None:
            raise ToriiException('Cluster is not defined on {0}'.format(self))

        if self._cluster_obj is None or self._cluster_obj.id != self.cluster.id:
            self._cluster_obj = self._service._torii.clusters.get(self.cluster.id)

        return self._cluster_obj

    @property
    def sched_options(self):
        if not hasattr(self, 'schedOptions') or self.schedOptions is None:
            self.schedOptions = []

        return self.schedOptions


    def _get_sched_option(self, name):
        """ Get the sched_option object based on its name """
        option = None

        # get the option from the task
        if self.sched_options:
            option = next((option for option in self.sched_options
                           if option.label.lower().strip() == name.lower().strip()), None)

        return option


    def clear_sched_options(self, name):
        """
        Clear the value or the file list of an sched_option

        :param name: the name of the sched_option
        :return:
        """
        self.schedOptions = [option for option in self.sched_options
                             if option.label.lower().strip() != name.lower().strip()]


    def set_sched_option(self, name, value=None):
        """
        Set an sched_option value based on its name

        :param name: the name of the sched_option
        :param value: the value to set
        """

        # get the option from the task if it already exist
        option = self._get_sched_option(name)

        # get the option from the cluster if not
        if option is None and self.cluster_obj is not None:
            option = self.cluster_obj.get_sched_option(name)

            if option is None:
                raise ToriiException('No scheduler option {0} in {1}'.format(name, self.cluster_obj))

            # add a copy of the option to the task
            option = copy.deepcopy(option)
            self.sched_options.append(option)

        # careful value attribute may not exist
        setattr(option, 'value', value)

    def get_sched_option(self, name):
        """
        Get an scheduler option value based on its name

        :param name: the name of the sched_option 
        :return: the value
        """
        option = self._get_sched_option(name)

        if option is not None:
            return option.value
        else:
            return None

    @property
    def cluster(self):
        if hasattr(self, 'context') and hasattr(self.context, 'cluster'):
            return self.context.cluster
        else:
            return None

    def set_cluster(self, cluster):
        if isinstance(cluster, ToriiObject):
            setattr(self.context, 'cluster', cluster.ref)
        elif isinstance(cluster, Struct):
            setattr(self.context, 'cluster', cluster)
        else:
            raise ToriiException('Bad parameter, expecting a ToriiObject or a Reference')

    def set_rundir(self, rundir):
        setattr(self.context, 'runDir', rundir)

    def submit(self):
        return self._service.submit(self)

    def lowla(self, **kwargs):
        return self._service.lowla(self, **kwargs)

    def validate(self):
        return self._service.validate(self)

    def clone(self):
        return self._service.clone(self)

    def picomize(self):
        return self._service.picomize(self)

    def kill(self):
        return self._service.kill(self)

    def pause(self):
        return self._service.pause(self)

    def resume(self):
        return self._service.resume(self)

    def command(self, commandId, commandOption=None):
        return  self._service.command(self, commandId=commandId, commandOption=commandOption)

    def job_command(self, jobIdx, commandId, commandOption=None):
        return  self._service.job_command(self, jobIdx=jobIdx, commandId=commandId, commandOption=commandOption)

    def get_records(self, max_records=100, increment=True):
        return self._service.get_records(self, max_records, increment)

    @property
    def viewer_url(self):
        return self._service.get_viewer_url(self)

    def open_viewer(self):
        self._service.open_viewer(self)

    @property
    def mapper_url(self):
        return self._service.get_mapper_url(self)

    def open_container_mapper(self):
        self._service.open_container_mapper(self)



from torii.dao.business_object_dao import BusinessObjectDao


class DaoManager:
    """
    The manager for all data access objects (DAOs).

    All DAOs should be mapped to the 'services' dictionary, where the Key is the
    first element in the REST API path and the value the uninitialized DAO class.
    """

    services = {
        'bo': BusinessObjectDao,
    }

    def __init__(self, torii):
        self._torii = torii
        self._logger = self._torii.logger

        for key, bo in self.services.items():
            # Check if bo is a class and not an instance
            if isinstance(bo, type):
                self.services[key] = bo(torii)

    def run_service(self, intent: str, path: str, params=None, json=None, data=None, files=None, timeout=None):
        """
        Runs the intended service corresponding to the given intent and path.
        Passes the defined arguments to the service.

        :param intent: The Http method (GET, POST, PUT, DELETE or PATCH).
        :param path: The REST API path (ex.: bo/{boClass}/page).
        :param params: Optional keyword argument.
        :param json: Optional argument.
        :param data: Optional argument.
        :param files: Optional argument.
        :param timeout: Optional argument.
        :returns: The class method if exists, otherwise None.
        """

        if not self._torii.properties['use_python_dao']:
            return None

        path = path.split('/')
        service = path.pop(0)

        if service not in self.services:
            return None

        args = [intent, '/'.join(path)]
        if json: args.append(json)
        if data: args.append(data)
        if files: args.append(files)
        if timeout: args.append(timeout)
        kwargs = params or {}

        try:
            return self.services[service].init(*args, **kwargs)
        except Exception as e:
            self._logger.info('Error in DAO Manager run_service:', e)
            raise e



import bson
import os
import sys

from torii.services.service import Service
from torii.data.torii_object import ToriiObject
from torii.exception import ToriiException

class FileService(Service):

    def __init__(self, torii):
        """
        Create a generic Gateway service
        """

        Service.__init__(self, torii, "files")

    def download(self, server, path, fd, chunk_size=65536):
        """
        Download a file from the server

        :param server: the server to download from
        :param path: the absolute path on this server
        :param fd: the file descriptor to write on
        :param chunk_size: the chunk size for the streaming
        """
        if not isinstance(fd, file):
            raise ToriiException('Bad parameters, fd is mandatory')

        if not (isinstance(path, str) and path.startswith('/')):
            raise ToriiException('Bad parameters, path is mandatory and must be an absolute path (starts with /)')

        params = {'absolutePath' : path,
                  'fileName': 'default'}

        if isinstance(server, ToriiObject):
            params['serverId'] = server.id
        elif bson.objectid.ObjectId.is_valid(server):
            params['serverId'] = server
        elif isinstance(server, str):
            params['serverName'] = server
        else:
            raise ToriiException('Bad parameters, server is mandatory and must be a server object or the server name')

        try:
            r = self._torii.http_session.request(method='GET',
                                                 url=self.get_url('download'),
                                                 stream=True,
                                                 params=params)

            if r.status_code != 200:
                raise ToriiException('Failed to download {0} on {1} (code = {3})\n{2}'
                                     .format(path, server, r.reason, r.status_code))

            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk: # filter out keep-alive new chunks
                    fd.write(chunk)
                    fd.parse()
                    os.fsync(fd.fileno())

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to download {0} on {1} \n{2}'.format(path, server, e)).with_traceback(sys.exc_info()[2])


    def upload(self, server, path, fd):
        """
        Upload a file to the server

        :param server: the server to upload to
        :param path: the absolute path on this server (for the file or the destination directory)
        :param fd: the file descriptor to read from
        """

        if not isinstance(fd, file):
            raise ToriiException('Bad parameters, fd is mandatory')

        if not (isinstance(path, str) and path.startswith('/')):
            raise ToriiException('Bad parameters, path is mandatory and must be an absolute path (starts with /)')

        # path can contain be interpreted as a directory path or a file path
        # it depends if it ends with '/' or not
        if path.endswith('/'):
            abs_path = path[:-1]
            filename = fd.name.split('/')[-1]
        else:
            path_split = path.split('/')
            abs_path = '/'.join(path_split[0:-1]) if len(path_split) > 2 else '/{0}'.format(path_split[0])
            filename = path.split('/')[-1]

        params = {'absolutePath' : abs_path,
                  'filename': filename}

        if isinstance(server, ToriiObject):
            params['serverId'] = server.id
        elif bson.objectid.ObjectId.is_valid(server):
            params['serverId'] = server
        elif isinstance(server, str):
            params['serverName'] = server
        else:
            raise ToriiException('Bad parameters, server is mandatory and must be a server object or the server name')
        
        try:
            r = self._torii.http_session.request(method='POST',
                                                 headers={'Content-Type': 'application/octet-stream'},
                                                 url=self.get_url('direct_upload'),
                                                 params=params,
                                                 stream=True,
                                                 data=fd)
            if r.status_code not in [200, 204]:
                raise ToriiException('Failed to upload {0} from {1} (code = {2}) \n{3}'
                                     .format(path, server, r.status_code, r.reason))

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to upload {0} from {1} \n{2}'.format(path, server, e)).with_traceback(sys.exc_info()[2])




import torii
import pydoc
import os
import pkgutil

src_dir = os.path.dirname(os.path.realpath(__file__))
docs_dir = os.path.join(src_dir, 'docs')

if not os.path.exists(docs_dir):
    os.makedirs(docs_dir)

os.chdir(docs_dir)
package = torii
# pydoc.writedocs(src_dir)
pydoc.writedoc('torii')
for importer, modname, ispkg in pkgutil.walk_packages([src_dir], ''):
    if not modname.startswith('tests') and modname not in ['generate_doc', 'torii_main']:
        pydoc.writedoc('torii.' + modname)



from time import time

from requests import Session
from requests.adapters import HTTPAdapter


class CustomSession(Session):

    def __init__(self):
        super().__init__()
        self.t = time()

    def reset_connection(self):
        self.close()
        adapter = HTTPAdapter()
        self.mount('https://', adapter)
        self.mount('http://', adapter)

    def request(self, *args, **kwargs):
        # If it has been more than 4 minutes since last request
        if (time() - self.t) >= 240:
            self.reset_connection()
        self.t = time()

        response = super().request(*args, **kwargs)
        response.raise_for_status()

        return response



import os
from time import sleep

from requests_toolbelt import MultipartEncoder
import json as json_lib

from torii.data import *
from torii.services import Service


class AiService(Service):

    def __init__(self, torii):
        """
        Create the graph service
        """
        Service.__init__(self,
                         torii=torii,
                         name='ai',
                         base_path='artificial_intelligence_c9a0d5e4')



    def start_engine(self,
                     application_name, application_version,
                     model_path, model_server,
                     task_name=None, tags=None):

        payload = {
            'type': 'inferenceInputs',
            'applicationName': application_name,
            'applicationVersion': application_version,
            'modelPath': model_path,
            'serverName': model_server.name if isinstance(model_server, Struct) else model_server,
            'taskName': task_name if task_name else application_name,
            'tags': tags if tags else [],
            'config' : {
                'one_shot': False,
                'input_server': model_server.name if isinstance(model_server, Struct) else model_server
            }
        }

        json = self.request_post(path='inference', json=payload)

        ai_task = Struct(json)
        ai_task.task = Task(json=json['task'], service=self._torii.tasks)

        return ai_task

    def init_proxy(self, engine):

        # Map<String, Object> params = new HashMap<>();
        # params.put("https", false);
        # params.put("port", inferenceEngine.getTask().getRunInfos().get(InferenceEngine.HTTP_PORT_KEY));
        # params.put("path", "/");
        #
        # try {
        #     return this.getRequest(inferenceEngine.getTask().getIdString() + "/" + PROXY_TARGET,
        #             params,
        #             InferenceEngine.class);
        # } catch (Exception e) {
        #     throw new ToriiException("Failed to create proxy for inference engine " + inferenceEngine.getId(), e);
        # }

        sleep(3)
        engine.task.update()
        params = {
            'https': False,
            'port': engine.task.runInfos.http_port,
            'path': '/'
        }

        self.request_get(path='inference/{}/proxy'.format(engine.task.id), params=params)
        engine.task.update()


    def infer(self, engine, images, config=None):


        if not engine.task.proxies:
            self.init_proxy(engine)

        proxy = engine.task.proxies[0]

        if config is None:
            config = {
                'type': 'jpg',
                'confidence_threshold': 0.1
            }

        multipart_data = {'config': json_lib.dumps(config)}
        multipart_data = {}

        i = 0
        for image_path in images:
            key = 'file' + str(i) + os.path.splitext(image_path)[1]
            multipart_data[key] = (os.path.basename(image_path),
                                   open(image_path, 'rb'),
                                   'application/octet-stream')

            i += 1

        multipart_encoder = MultipartEncoder(fields=multipart_data)
        result = self.request_post(path='/proxy/{}/'.format(proxy.id),
                                   data=multipart_encoder,
                                   params=config)

        return Struct(result)



from torii.data import *
from torii.services.torii_service import ToriiService


class OrganisationService(ToriiService):

    def __init__(self, torii):
        """
        Create the picom service
        """
        ToriiService.__init__(self, torii=torii, name='organisations', object_class=ToriiObject)


    @property
    def local(self):
        """
        Get the local organisation
        :return:
        """

        return ToriiObject(self.request_get('local'))



from torii.data import *
from torii.services.torii_service import ToriiService


class BSService(ToriiService):

    def __init__(self, torii):
        """
        Create the BO service

        :param torii:
        :param bo_name:
        """
        ToriiService.__init__(self, torii=torii, name='pythonServices', object_class=ToriiObject)

    def tail(self, bs, limit=100):
        """
        Start a business service

        :param bs: the business service
        :return:
        """

        json = self.request('GET', path='{0}/tail'.format(bs.id), params={'limit': limit},)
        bs.from_json(json)

        return bs

    def start(self, bs):
        """
        Start a business service

        :param bs: the business service
        :return:
        """

        json = self.request('GET', path='{0}/start'.format(bs.id))
        bs.from_json(json)

        return bs

    def stop(self, bs):
        """
        Stop a business service

        :param bs: the business service
        :return:
        """

        json = self.request('GET', path='{0}/stop'.format(bs.id))
        bs.from_json(json)

        return bs

    def restart(self, bs):
        """
        Restart a business service

        :param bs: the business service
        :return:
        """

        json = self.request('GET', path='{0}/restart'.format(bs.id))
        bs.from_json(json)

        return bs

    def status(self, bs=None):
        """
        Get the status of one/all business service

        :param bs: the business service(s)
        :return:
        """

        if bs:
            json = self.request('GET', path='{0}/status'.format(bs.id))
            bs.from_json(json)

            return bs
        else:
            json = self.request('GET', path='status')

            bss = [self._object_class(x, service=self) for x in json]
            return bss
